/*!
  \file   parms_mat_impl.h
  \brief  The matrix structure and functions

  \author zzli
  \date   2006-05-08
*/

#ifndef _PARMS_MAT_IMPL_H_
#define _PARMS_MAT_IMPL_H_

#include "parms_mat.h"
#include "parms_operator.h"
#include "parms_map_impl.h"
#include "parms_vec_impl.h"


extern o_func mat_tab[];

#define PARMS_MAT_FORMAT_KIND    0x00f00000
#define PARMS_MAT_SHIFT          20
#define PARMS_MAT_GET_KIND(a)       (((a)&PARMS_MAT_FORMAT_KIND) >> PARMS_MAT_SHIFT)
#define PARMS_MAT_SET_KIND(a, kind) ((a)|(kind) << PARMS_MAT_SHIFT)

typedef  int (*PARMS_ILU)(parms_Mat self, parms_FactParam param, void
			  *data, parms_Operator *op);  

/*! \struct parms_Mat_ops.
  \brief struct parms_mat_ops.
 */
typedef struct parms_Mat_ops {
  /*! function parameter: apply
    \brief performs \f$y = self \times x\f$. 
  */
  int (*apply)(parms_Mat self, parms_Vec x, parms_Vec y);
  /*! 
    Divide the local equation into diagonal part and off-diagonal
    part. Set up communcation handler for matrix-vector product.
   */
  int (*setup)(parms_Mat self);	
  /*! Set the communication type
    \f{tabular}{ll}
    P2P & Copying data into a temporary buffer \\
    DERIVED & Creating a derived datatype as in the old version of
    pARMS. 
    \f}
   */
  int (*setcommtype)(parms_Mat self, COMMTYPE ctype);

  /*! function parameter: mvpy
    \brief Performs z = beta*y + alpha*self*x.
  */
  int (*mvpy)(parms_Mat self, REAL alpha, parms_Vec x, REAL beta,
	      parms_Vec y, parms_Vec z);

  /*! function parameter: getdiag
    \brief Get the diagonal part of the local matrix.
   */
  int (*getdiag)(parms_Mat self, void **mat);
  /*! function parameter: getlmat
    \brief Get the local matrix including diagonal and off-diagonal
    matrix.  
   */
  int (*getlmat)(parms_Mat self, void **mat);
  /*! function parameter: extend
    \brief Extend the submatrix mat by including equations that
    correspond to immediate neighbouring variables.
   */
  int (*extend)(parms_Mat self, parms_Comm handler, int start, void
		*mat, int *n, void **ext_mat);
  /*! function parameter: setilu
    \ Description: Set ILU type. Supported ILU type:

    -PCILU0
    -PCILUK
    -PCILUT
    -PCARMS
   */
  int (*setilu)(parms_Mat self, PCILUTYPE pcstype);
  /*! function parameter: ilu
    \brief ILU factorization. 
   */
  int (*ilu)(parms_Mat self, parms_FactParam param, void *data,
	     parms_Operator *op);
  /*! function Parameter: mvoffd
    \brief The matrix-vector product for off-diagonal part.
   */
  int (*mvoffd)(parms_Mat self, FLOAT *x, FLOAT *y, int pos);

  /*! function Parameter: matfree
    \brief Free the matrix mat.
   */
  int (*matfree)(parms_Mat self, void *mat);

  /*! function Parameter: gethandler
      \brief Get the communication handler for mv product.
   */
  int (*gethandler)(parms_Mat self, parms_Comm *handler);
} *parms_Mat_ops;


/*! \struct parms_vcsr.
  \ Description: struct parms_vcsr.
 */
typedef struct parms_vcsr {
  int      n;         	//!< the dimension of the matrix 
  int      *nnzrow;    //!< the length of each row 
  int      *space;     //!<  length of space 
  /*! 
  		Parameter: pj = An indirect pointer to store column indices.
   */
  int      **pj;
  /*! parameter: pa = An indirect pointer to store corresponding nonzero entries.  
  */
  FLOAT    **pa;
} *parms_vcsr;

/*! \struct parms_Mat_
  \Description: struct parms_Mat_
 */
struct parms_Mat_ {
  PARMS_OBJHEADER(parms_Mat_ops)
  void        *data;
  BOOL        isserial;
  BOOL        issetup;
  BOOL        isperm;
  BOOL        isalloc;
  MATTYPE     type;
  int         m,n;
  int         M,N;
  parms_Map   is;
  parms_vcsr  aux_data;
};

#endif 
