/*!
  \file   parms_object_impl.h
  \brief  parms_Object is the parent of all the other objects in
          pARMS.  

  \author zzli
  \date   2006-05-05
*/

#ifndef _PARMS_OBJECT_IMPL_H_
#define _PARMS_OBJECT_IMPL_H_

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "parms_sys.h"
#include "parms_mem.h"
#include "parms_viewer.h"

#define PARMS_OBJECT_CLASS    0xf0000000
#define PARMS_MAP_CLASS       0x0
#define PARMS_VEC_CLASS       0x1
#define PARMS_MAT_CLASS       0x2
#define PARMS_PC_CLASS        0x3
#define PARMS_SOLVER_CLASS    0x4
#define PARMS_OPERATOR_CLASS  0x5
#define PARMS_TIMER_CLASS     0x6
#define PARMS_VIEWER_CLASS    0x7
#define PARMS_COMM_CLASS      0x8

#define PARMS_OBJECT_MASK     0x01000000
#define PARMS_OBJECT_SHIFT    28
#define PARMS_OBJECT_GET_CLASS(a)    (((a)&PARMS_OBJECT_CLASS) >> PARMS_OBJECT_SHIFT)
#define PARMS_OBJECT_SET_CLASS(kind) (PARMS_OBJECT_MASK|(kind) << PARMS_OBJECT_SHIFT)

#if defined(GCC3)
char errbuf[100];
#define MAKESTRING(msg) sprintf(errbuf, "%s%s", #msg, " at line %d in file %s\n")
#define ERRORMSG(msg) (MAKESTRING(msg), fprintf(stderr, errbuf, __LINE__, __FILE__))
#else
#define MAKESTRING(msg) #msg
#define CONCATMSG(msg)  MAKESTRING(msg) ## " at line %d in file %s\n"
#define ERRORMSG(msg) fprintf(stderr, CONCATMSG(msg), __LINE__, __FILE__)
#endif 

#ifdef USE_MPI
#define PARMS_ABORT(error_number)			\
  if (true) {						\
    int flag;						\
    MPI_Initialized(&flag);				\
    if (flag) {						\
      MPI_Abort(MPI_COMM_WORLD, (error_number));	\
    }							\
    else {						\
      exit((error_number));				\
    }							\
  } else
#else
#define PARMS_ABORT(error_number)		\
  if (true) {					\
    exit((error_number));			\
  } else
#endif

typedef struct parms_Object *parms_Object;
typedef int (*o_dtor)(parms_Object *);
typedef int (*o_func)(parms_Object);
typedef int (*o_view)(parms_Object, parms_Viewer);

/*!  \def PARMS_OBJGETREF(o)
     \brief Get the reference number of object \a o.
 */
#define PARMS_OBJGETREF(o)  (o)->ref
/*! \def  PARMS_OBJADDREF(o)
    \brief Increase the reference number of object \a o.
 */
#define PARMS_OBJADDREF(o)			\
  if (true) {					\
    PARMS_VALIDPTR((o));			\
    (o)->ref++;					\
  } else

/*! \def  PARMS_OBJSUBREF(o) 
    \brief Decrease the reference number of object \a o.
 */
#define PARMS_OBJSUBREF(o)			\
  if (true) {					\
    PARMS_VALIDPTR((o));			\
    (o)->ref--;					\
  } else

/*! \struct parms_object_vtbl
  \brief  function pointer table of parms_Object which inherited by
  all derived objects in pARMS.
 */
typedef struct parms_Object_vtbl {
  /*! \typedef a function pointer to the destructor.
   */
  o_dtor          dtor;
  /*! \typedef a fucntion pointer to the function dumping object.
   */
  o_view          view;
} *parms_Object_vtbl;


#define PARMS_OBJHEADER(DVPTR)			\
  int               handler;			\
  parms_Object_vtbl o_vptr;			\
  DVPTR             d_vptr;			\
  int               ref;  


struct parms_Object {
  PARMS_OBJHEADER(int *)
};

/*! \def  PARMS_OBJCREATE(o, o_dtor, o_view)
    \brief Create parms_Object \a o and fill in associated virtual
    function pointer table parms_object_vtbl with \a o_dtor and \a
    o_view. 
 */
#define PARMS_OBJCREATE(o, o_dtor, o_view)	 \
  if (true) {					 \
    PARMS_NEW0((o));				 \
    PARMS_NEW0((o)->o_vptr);			 \
    PARMS_NEW0((o)->d_vptr);			 \
    (o)->ref = 1;				 \
    (o)->o_vptr->dtor  = (o_dtor);		 \
    (o)->o_vptr->view  = (o_view);		 \
  } else

/*! \def PARMS_OBJFREE(o)
    \brief Free object object \a o.
 */
#define PARMS_OBJFREE(o)			\
  if (true) {					\
    PARMS_FREE((o)->o_vptr);			\
    PARMS_FREE((o)->d_vptr);			\
    PARMS_FREE((o));				\
    (o) = NULL;					\
  } else 


#define PARMS_VALIDPTR(h)			\
  if (true) {					\
    assert((h));				\
    if (!((h)->handler & PARMS_OBJECT_MASK)) {	\
      ERRORMSG(Invalid pointer);		\
      PARMS_ABORT(66);				\
    }						\
  } else

#define PARMS_VALIDMAPPTR(h)						\
  if (true) {								\
    parms_Object o;							\
    o = (parms_Object)(h);						\
    PARMS_VALIDPTR(o);							\
    if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_MAP_CLASS) {	\
      ERRORMSG(Invalid parms_Map object);				\
      PARMS_ABORT(67);							\
    }									\
  } else

#define PARMS_VALIDVECPTR(h)						\
  if (true) {								\
    parms_Object o;							\
    o = (parms_Object)(h);						\
    PARMS_VALIDPTR(o);							\
    if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_VEC_CLASS) {	\
      ERRORMSG(Invalid parms_Vec object);				\
      PARMS_ABORT(68);							\
    }									\
  } else

#define PARMS_VALIDMATPTR(h)						\
  if (true) {								\
    parms_Object o;							\
    o = (parms_Object)(h);						\
    PARMS_VALIDPTR(o);							\
    if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_MAT_CLASS) {	\
      ERRORMSG(Invalid parms_Mat object);				\
      PARMS_ABORT(69);							\
    }									\
  } else

#define PARMS_VALIDPCPTR(h)					\
  if (true) {							\
    parms_Object o;						\
    o = (parms_Object)(h);					\
    PARMS_VALIDPTR(o);						\
    if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_PC_CLASS) {	\
      ERRORMSG(Invalid parms_PC object);			\
      PARMS_ABORT(70);						\
    }								\
  } else

#define PARMS_VALIDSOLVERPTR(h)						\
  if (true) {								\
    parms_Object o;							\
    o = (parms_Object)(h);						\
    PARMS_VALIDPTR(o);							\
    if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_SOLVER_CLASS) {	\
      ERRORMSG(Invalid parms_Solver object);				\
      PARMS_ABORT(71);							\
    }									\
  } else

#define PARMS_VALIDOPTPTR(h)						\
  if (true) {								\
    parms_Object o;							\
    o = (parms_Object)(h);						\
  PARMS_VALIDPTR(o);							\
  if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_OPERATOR_CLASS) {	\
    ERRORMSG(Invalid parms_Operator object);				\
    PARMS_ABORT(72);							\
  }									\
 } else

#define PARMS_VALIDTIMERPTR(h)						\
  if (true) {								\
    parms_Object o;							\
    o = (parms_Object)(h);						\
    PARMS_VALIDPTR(o);							\
    if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_TIMER_CLASS) {	\
      ERRORMSG(Invalid parms_Timer object);				\
      PARMS_ABORT(73);							\
    }									\
  } else 

#define PARMS_VALIDVIEWERPTR(h)						\
  if (true) {								\
    parms_Object o;							\
    o = (parms_Object)(h);						\
    PARMS_VALIDPTR(o);							\
    if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_VIEWER_CLASS) { 	\
      ERRORMSG(Invalid parms_Viewer object);				\
      PARMS_ABORT(74);							\
    }									\
  } else

#define PARMS_VALIDCOMMPTR(h)						\
  if (true) {								\
    parms_Object o;							\
    o = (parms_Object)(h);						\
    PARMS_VALIDPTR(o);							\
    if (PARMS_OBJECT_GET_CLASS(o->handler) != PARMS_COMM_CLASS) { 	\
      ERRORMSG(Invalid parms_Viewer object);				\
      PARMS_ABORT(74);							\
    }									\
  } else

#endif
