/**
 * @file   parms_solver.h
 * @author Zhongze Li
 * @date   Tue Oct 17 11:58:22 2006
 * 
 * @brief  Functions related to the Krylov subspace methods. Only
 *         FGMRES is supported. 
 * 
 */

#ifndef _PARMS_SOLVER_H_
#define _PARMS_SOLVER_H_

#include "parms_vec.h"
#include "parms_mat.h"
#include "parms_viewer.h"
#include "parms_pc.h"
#include "parms_operator.h"

PARMS_CXX_BEGIN

typedef struct parms_Solver_ *parms_Solver;

/** 
 * Setup parms_Solver solver self.
 * 
 * @param self A parms_Solver object.
 * 
 * @return 0 on success.
 */
extern int parms_SolverSetup(parms_Solver self);

/** 
 * Get the matrix of the linear system.
 * 
 * @param self A parms_Solver object.              
 * @param A    A pointer to the matrix returned.   
 * 
 * @return 0 on success.
 */
extern int parms_SolverGetMat(parms_Solver self, parms_Mat *A);

/** 
 * Get the preconditioning matrix.
 * 
 * @param self A parms_Solver object.                    
 * @param PC   A pointer to the preconditioning matrix.  
 * 
 * @return 0 on success.
 */
extern int parms_SolverGetPC(parms_Solver self, parms_PC *PC);

/** 
 * Solve the equation \f$Ax = y\f$.
 * 
 * @param self A parms_Solver object.     
 * @param x    The solution vector.       
 * @param y    The right-hand-side vector.
 * 
 * @return 0 on success.
 */
extern int parms_SolverApply(parms_Solver self, parms_Vec x, parms_Vec
			     y); 

/** 
 * Set the type of the solver.
 *
 * Only FGMRES solver is available in the package.
 * 
 * @param self   A parms_Solver object.       
 * @param stype  The type of Krylov subspace. 
 *               - SOLFGMRES
 *               - SOLDGMRES
 *               
 * @return 0 on success.
 */
extern int parms_SolverSetType(parms_Solver self, SOLVERTYPE stype);

/** 
 * Create a parms_Solver object.
 * 
 * @param self A pointer to the parms_Solver object created.
 * @param A    The matrix of the linear system.   
 * @param pc   The preconditioner.  
 * 
 * @return 0 on success.
 */
extern int parms_SolverCreate(parms_Solver *self, parms_Mat A, parms_PC pc);

/** 
 * Free the memory for the parms_Solver object.
 * 
 * @param self A pointer to the parms_Solver object to be freed. 
 * 
 * @return 0 on success.
 */
extern int parms_SolverFree(parms_Solver *self);

/** 
 * Dump te solver object.
 * 
 * @param self A parms_Solver object.   
 * @param v    A parms_Viewer object.   
 */
extern void parms_SolverView(parms_Solver self, parms_Viewer v);

/** 
 * Set parameter for the solver.
 * 
 * Set the maximum iteration counts, the restart size of GMRES, and
 * the convergence tolerance.
 * 
 * @param self   A parms_Solver object.
 * @param ptype  The type of parameter.
 *               - MAXITS maximum iteration counts.
 *               - KSIZE  restart size of GMRES.
 *               - DTOL   converence tolerance.
 *               - NEIG   number of eigenvectors.
 * @param param  Parameters for the solver.
 */
extern void parms_SolverSetParam(parms_Solver self, PARAMTYPE ptype,
				 char *param);
   
/** 
 * Get the iteration counts.
 * 
 * @param self A parms_Solver object.
 * 
 * @return The iteration counts.
 */
extern int parms_SolverGetIts(parms_Solver self);

PARMS_CXX_END

#endif 
