/**
 * @file   parms_sys.h
 * @author Zhongze Li
 * @date   Tue Oct 17 11:59:09 2006
 * 
 * @brief  Macros and typedef needed by all other header files. 
 * 
 */

#ifndef _PARMS_SYS_H_
#define _PARMS_SYS_H_

#include <stdio.h>
#ifdef C99
#include <stdbool.h>
#endif 
#include "mpi.h"

#if defined(__cplusplus)
#define PARMS_CXX_BEGIN extern "C" {
#define PARMS_CXX_END   }
#else
#define PARMS_CXX_BEGIN
#define PARMS_CXX_END
#endif 

PARMS_CXX_BEGIN

typedef enum VARSTYPE   {INTERLACED, NONINTERLACED} VARSTYPE;
typedef enum INSERTMODE {INSERT, ADD} INSERTMODE;
typedef enum COMMTYPE   {P2P, DERIVED} COMMTYPE;
typedef enum SOLVERTYPE {SOLFGMRES, SOLDGMRES} SOLVERTYPE;
typedef enum PARAMTYPE  {MAXITS, KSIZE, DTOL, NEIG} PARAMTYPE;
typedef enum MATTYPE    {MAT_NULL=-1, MAT_VCSR=0, MAT_CSR=1} MATTYPE;
typedef enum PCTYPE     {PCBJ, PCSCHUR, PCRAS} PCTYPE;
typedef enum PCILUTYPE  {PCILU0, PCILUK, PCILUT, PCARMS} PCILUTYPE;

#ifdef C99
typedef _Bool BOOL;
#else
typedef  int  BOOL;
#define true  1
#define false 0
#endif 

#define PARMS_COUT "parms_cout"
#define PARMS_CERR "parms_cerr"

#define FINDEX 1

typedef struct parms_FactParam {
  int    schur_start;		/* start position of the Schur complement */
  int    start;			/* start row index of the submatrix in
				   the whole matrix */
  int    n;			/* the size of the local matrix */
  /*
    mc = multicoloring or not (0-not, 1-yes).
    ipar[0:17]  = integer array to store parameters for both
       arms construction (arms2) and iteration (armsol2).
    
       ipar[0]:=nlev.  number of levels (reduction processes).  see
       also "on return" below.
    
       ipar[1]:=bsize. Dimension of the blocks. In this version, bsize
       is only a target block size. The size of each block can vary
       and is >= bsize.
    
       ipar[2]:=iout if (iout > 0) statistics on the run are printed
       to FILE *ft

       The following are not used by arms2 -- but should set before
       calling the preconditionning operation armsol2:

       ipar[3]:= Krylov subspace dimension for last level 
         ipar[3] == 0 means only backward/forward solve is performed
            on last level.                   
       ipar[4]:=  maximum # iterations on last level
    
       ipar[5-9] NOT used [reserved for later use] - must be 
       set to zero. [see however a special use of ipar[5] in fgmresC.] 
    
       The following set method options for arms2. Their default
       values can all be set to zero if desired.  
    
       ipar[10-13] == meth[0:3] = method flags for interlevel blocks
       ipar[14-17] == meth[0:3] = method flags for last level block -
         with the following meaning  
           meth[0] permutations of rows  0:no 1: yes. affects rperm
             NOT USED IN THIS VERSION ** enter 0.. Data: rperm
           meth[1] permutations of columns 0:no 1: yes. So far this is 
             USED ONLY FOR LAST BLOCK [ILUTP instead of ILUT]. (so
             ipar[11] does no matter - enter zero). If ipar[15] is one
             then ILUTP will be used instead of ILUT. Permutation data
             stored in: perm2.  
           meth[2] diag. row scaling. 0:no 1:yes. Data: D1
           meth[3] diag. column scaling. 0:no 1:yes. Data: D2
             similarly for meth[14], ..., meth[17] all transformations
             related to parametres in meth[*] (permutation,
             scaling,..) are applied to the matrix before processing
             it   
    
    droptol  = Threshold parameters for dropping elements in ILU
       factorization. 
      droptol[0:4] = related to the multilevel  block factorization
      droptol[5:5] = related to ILU factorization of last block.
        This flexibility is more than is really needed. one can use a
        single parameter for all. it is preferable to use one value
        for droptol[0:4] and another (smaller) for droptol[5:6] 
      droptol[0] = threshold for dropping  in L [B]. See piluNEW.c:
      droptol[1] = threshold for dropping  in U [B].
      droptol[2] = threshold for dropping  in L^{-1} F 
      droptol[3] = threshold for dropping  in E U^{-1} 
      droptol[4] = threshold for dropping  in Schur complement
      droptol[5] = threshold for dropping  in L in last block [see
        ilutpC.c] 
      droptol[6] = threshold for dropping  in U in last block [see
        ilutpC.c] 
    
    lfil     = lfil[0:6] is an array containing the fill-in parameters.
      similar explanations as above, namely:
      lfil[0] = amount of fill-in kept  in L [B]. 
      lfil[1] = amount of fill-in kept  in U [B].
      etc..
    
    tolind   = tolerance parameter used by the indset function. 
      a row is not accepted into the independent set if the *relative*
      diagonal tolerance is below tolind. see indset function for
      details. Good values are between 0.05 and 0.5 -- larger values
      tend to be better for harder problems.
    
    Note:   The comments above are for arms. The first element for every 
    array is for other preconditioner: ilut, iluk
  */
  int    mc;		
  int    lfil[7];	
  int    ipar[18];	
  double droptol[7];
  double pgfpar[2];
  double tolind;
  BOOL   isalloc;
} *parms_FactParam;

#define ZERO      0.0
#define EPSILON   1.0e-20
#define EPSMAC    1.0e-16  

/* external BLAS */
#if defined(DBL)
#define FLOAT   double
#define GDOT    DDOT
#define GCOPY   DCOPY
#define GSCAL   DSCAL
#define GAXPY   DAXPY
#define GNRM2   DNRM2
#define GDMIN   IDMIN
#define GGEMV   DGEMV
#define GGEMM   DGEMM
#define GGETRF  DGETRF
#define GGETRI  DGETRI
#define GGESVD  DGESVD
#if defined(FORTRAN_CAPS)
#define ddot_     DDOT
#define dcopy_    DCOPY
#define dscal_    DSCAL
#define daxpy_    DAXPY
#define dnrm2_    DNRM2
#define idmin_    IDMIN
#define dgemv_    DGEMV
#define dgemm_    DGEMM
#define dgetrf_   DGETRF
#define dgetri_   DGETRI
#define dgesvd_   DGESVD
#elif defined(FORTRAN_DOUBLE_UNDERSCORE)
#define ddot_     ddot__
#define dcopy_    dcopy__
#define dscal_    dscal__
#define daxpy_    daxpy__
#define dnrm2_    dnrm2__
#define idmin_    idmin__
#define dgemv_    dgemv__
#define dgemm_    dgemm__
#define dgetrf_   dgetrf__
#define dgetri_   dgetri__
#define dgesvd_   dgesvd__
#elif !defined(FORTRAN_UNDERSCORE)
#define BLAS_C_INTERFACE
#include <essl.h>
#define ddot_     ddot
#define dcopy_    dcopy
#define dscal_    dscal
#define daxpy_    daxpy
#define dnrm2_    dnrm2
#define idmin_    idmin
#define dgemv_    dgemv
#define dgemm_    dgemm
#define dgetrf_   dgetrf
#define dgetri_   dgetri
#define dgesvd_   dgesvd
#endif

/* FORTRAN subroutines */
#ifndef BLAS_C_INTERFACE
#define DDOT(n,x,incx,y,incy)        ddot_(&(n),(x),&(incx),(y),&(incy))
#define DCOPY(n,x,incx,y,incy)       dcopy_(&(n),(x),&(incx),(y),&(incy))
#define DSCAL(n,alpha,x,incx)        dscal_(&(n),&(alpha),(x), &(incx))
#define DAXPY(n,alpha,x,incx,y,incy) daxpy_(&(n), &(alpha), (x), \
					 &(incx), y, &(incy))
#define DNRM2(n, x, incx)            dnrm2_(&(n), (x), &(incx))
#define IDMIN(n, sx, incx)           idmin_((&(n), (sx), &(incx))
#define DGEMV(transa, m, n, alpha, a, lda, x, incx, beta, y, incy)  \
  dgemv_((transa), &(m), &(n), &(alpha), (a), &(lda), (x), &(incx), \
	 &(beta), (y), &(incy))
#define DGEMM(transa,transb,l,n,m,alpha,a,lda,b,ldb,beta,c,ldc)	\
  dgemm_((transa), (transb), &(l), &(n), &(m), &(alpha), (a),	\
	 &(lda), b, &(ldb), &(beta), (c), &(ldc)) 
#define DGETRF(m, n, a, lda, ipvt, info)		\
  dgetrf_(&(m), &(n), (a), &(lda), (ipvt), &(info))
#define DGETRI(n, a, lda, ipvt, work, lwork, info)			\
  dgetri_(&(n), (a), &(lda), (ipvt), (work), &(lwork), &(info))

extern double ddot_(int *n, double *x, int *incx, double *y, int
		    *incy);  
extern void   dcopy_(int *n, double *x, int *incx, double *y, int
		    *incy); 
extern void   dscal_(int *n, double *alpha, double *x, int *incx);
extern void   daxpy_(int *n, double *alpha, double *x, int *incx,
		    double *y, int *incy);
extern double dnrm2_(int *n, double *x, int *incx);
extern void   idmin_(int *n, double *sx, int *incx);
extern void   dgemv_(char *transa, int *m, int *n, double *alpha,
		     double *a, int *lda, double *x, int *incx, double
		     *beta, double *y, int *incy);
extern void   dgemm_(char *transa, char *transb, int *l, int *m, int
		     *n, double *alpha, double *a, int *lda, double
		     *b, int *ldb, double *beta, double *c, int *ldc);       
extern void   dgetrf_(int *m, int *n, double *a, int *lda, int *ipvt,
		      int *info); 
extern void   dgetri_(int *n, double *a, int *lda, int *ipvt, double
		      *work, int *lwork, int *info);
#else
#define DDOT(n,x,incx,y,incy)        ddot_((n), (x), (incx), (y), (incy)) 
#define DCOPY(n,x,incx,y,incy)       dcopy_((n), (x), (incx), (y), \
					   (incy)) 
#define DSCAL(n,alpha,x,incx)        dscal_((n), (alpha), (x), (incx)) 
#define DAXPY(n,alpha,x,incx,y,incy) daxpy_((n), (alpha), (x), (incx), \
					   (y), (incy)) 
#define DNRM2(n,x,incx)              dnrm2_((n), (x), (incx))

#define IDMIN(n,sx,incx)             idmin_((n), (sx), (incx))
#define DGEMV(transa,m,n,alpha,a,lda,x,incx,beta,y,incy)		\
  dgemv_((transa), (m), (n),						\
	 (alpha), (a), (lda), (x), (incx),				\
	 (beta), (y), (incy))

#define DGEMM(transa,transb,l,n,m,alpha,a,lda,b,ldb,beta,c,ldc)		\
  dgemm_((transa),(transb),						\
	 (l),(n),(m),(alpha),(a),					\
	 (lda),(b),(ldb),(beta),(c),(ldc))
#define DGETRF(m, n, a, lda, ipvt, info)  \
  dgetrf_((m), (n), (a), (lda), (ipvt), &(info))
#define DGETRI(n, a, lda, ipvt, work, lwork, info)		\
  dgetri_((n), (a), (lda), (ipvt), (work), (lwork), &(info))
#endif

extern void dgesvd_(char*, char*, int*, int*, double*, int*, double*,
		    double *, int*, double*, int*, double*, int*,
		    int*);     

#endif 

PARMS_CXX_END

#endif 
