/**
 * @file   parms_vec.h
 * @author Zhongze Li
 * @date   Tue Oct 17 12:01:25 2006
 * 
 * @brief  Functions related to the parms_Vec object. 
 * 
 * 
 */

#ifndef _PARMS_VECTOR_H_
#define _PARMS_VECTOR_H_

#include "parms_sys.h"
#include "parms_map.h"
#include "parms_viewer.h"
#include "parms_comm.h"

PARMS_CXX_BEGIN

typedef struct parms_Vec_ *parms_Vec;

/** 
 * Create a parms_Vec object.
 *
 * Create a parms_Vec object based on the data distribution indicated
 * by map.
 * 
 * @param self A pointer to the vector created.              
 * @param map  A map object which describes how variables are
 *             distributed across PEs. 
 * 
 * @return 0 on success.
 */
extern int parms_VecCreate(parms_Vec *self, parms_Map map);

/** 
 * Free the memory for the vector object.
 * 
 * @param self A pointer to the vector object.
 * 
 * @return 0 on success.
 */
extern int parms_VecFree(parms_Vec *self);

/** 
 * Dump the vector object.
 * 
 * Output the vector to the viewer v.
 * 
 * @param self A vector object.
 * @param v    A viewer object.
 * 
 * @return 0 on success.
 */
extern int parms_VecView(parms_Vec self, parms_Viewer v);

/** 
 * Get a pointer to a contiguous region containing the local part of
 * the distributed vector.
 * 
 * @param self  A vector object.                            
 * @param array A pointer to the local array in vector self.
 * 
 * @return 0 on success.
 */
extern int parms_VecGetArray(parms_Vec self, FLOAT **array);

/** 
 * Restore values to the parms_Vec object self.
 *
 * After calling this function, array cannot be used.
 * 
 * @param self   A vector object.                                  
 * @param array  A pointer to the array of values to be restored in
 *               the parms_Vec object self.         
 * 
 * @return 0 on success.
 */
extern int parms_VecRestoreArray(parms_Vec self, FLOAT **array);

/** 
 * Return the 2-norm of the vector.
 * 
 * @param self  A vector object.    
 * @param value The 2-norm returned.
 * 
 * @return 0 on success.
 */
extern int parms_VecGetNorm2(parms_Vec self, REAL *value);

/** 
 * Scale a vector.
 *
 * All components of vector self on the local processor times scalar. 
 * \f$self = scalar \times self\f$.
 *  
 * @param self   A vector object.
 * @param scalar A scalar.      
 * 
 * @return 0 on success.
 */
extern int parms_VecScale(parms_Vec self, FLOAT scalar);

/** 
 * Get the local size of parms_Vec object self.
 * 
 * @param self A vector object.
 * 
 * @return The local size of parms_Vec object self.
 */
extern int parms_VecGetLocalSize(parms_Vec self);

/** 
 * Get the global size of parms_Vec object self.
 * 
 * @param self A vector object.
 * 
 * @return The global size of parms_Vec object self.
 */
extern int parms_VecGetGlobalSize(parms_Vec self);

/** 
 * Set up the vector object self.
 *
 * This function must be called before the following parms_Vec
 * computation functions: parms_VecGetNorm2, parms_VecGetArray,
 * parms_VecRestoreArray, parms_VecScale, parms_VecAXPY,
 * parms_VecAYPX, parms_VecDOT, parms_VecDotArray.
 *  
 * @param self A parms_Vec object.
 * 
 * @return 0 on success.
 */
extern int parms_VecSetup(parms_Vec self);

/** 
 * Assign value to each component of the parms_Vec object self.
 * 
 * @param self   A vector object.         
 * @param value  A scalar value to be set.
 * 
 * @return 0 on success.
 */
extern int parms_VecPut(parms_Vec self, FLOAT value);

/** 
 * Insert values to parms_Vec object self.
 * 
 *  A pseudo code from the global point of view:
 *
 *  \f{verbatim}
 *  for (i = 0; i < m; i++) {
 *    self[im[i]] = values[i]; 
 *  }
 *  \f}
 *  
 * @param self   A vector object.                          
 * @param m      The number of variables to be inserted.   
 * @param im     An array of global variable indices.      
 * @param value  An array of values to be inserted to self.s 
 * @param mode   The style of set values:
 *               - ADD    add values to parms_Vec object self. 
 *               - INSERT assign values to parms_Vec object self.
 *               
 * @return 0 on success.
 */
extern int parms_VecSetValues(parms_Vec self, int m, int *im, FLOAT
			      *values, INSERTMODE mode); 

/** 
 * Perform \f$self := scalar \times x + self\f$.
 * 
 * @param self   A vector object.      
 * @param x      Another vector object.
 * @param scalar A scalar.
 * 
 * @return 0 on success.
 */
extern int parms_VecAXPY(parms_Vec self, parms_Vec x, FLOAT scalar);

/** 
 * Perform \f$self = scalar \times self + x\f$.
 * 
 * @param self    A vector object.      
 * @param x       Another vector object.
 * @param scalar  A scalar.
 * 
 * @return 0 on success.
 */
extern int parms_VecAYPX(parms_Vec self, parms_Vec x, FLOAT scalar);

/** 
 * Perform the inner product of two vectors.
 * 
 * If self and x are real vectors, value = self x^{T}. If self
 * and x are complex vectors, value = self \overline{x}^{T}.
 *
 * @param self   A vector object.                  
 * @param x      Another vector object.            
 * @param value  The inner product returned.       
 * 
 * @return 0 on success.
 */
extern int parms_VecDOT(parms_Vec self, parms_Vec x, REAL *value);

/** 
 * Perform the inner product between self and an array of parms_Vec
 * objects.  
 *
 * The pseudo code:
 *
 *  \f{verbatim}
 *  for (i = 0; i < n; i++) {
 *    result[i] = self * vecarray[i];
 *  }
 *  \f}
 *
 * @param self     A vector object.                              
 * @param n        The size of vecarray.                         
 * @param vecarray An array of vector objects.                   
 * @param aux      An auxiliary array.                           
 * @param result   An array of size n to store inner products.   
 * 
 * @return 0 on success.
 */
extern int parms_VecDotArray(parms_Vec self, int n, parms_Vec
			     *vecarray, FLOAT  *aux, FLOAT *result);   
/** 
 * Permute the vector object self.
 *
 * If the parms_Vec object and the parms_Mat object are created based
 * on  the same parms_Map object. Once matrix object is setup,
 * variables on each processor are divided into two cateries: internal
 * unknowns and interface unknowns. The parms_Vec should be permuted
 * accordingly. The user needn't call self function directly.
 * 
 * @param self A vector object.
 * 
 * @return 0 on success.
 */
extern int parms_VecPerm(parms_Vec self);

/** 
 * Attach a vector self to another vector x.
 *
 * Attach vector self to a chunk of vector x with beginning index
 * start. 
 *
 * self[0] \f$\rightarrow\f$  x[start] 
 *
 * self[1] \f$\rightarrow\f$  x[start+1]
 *
 * \f$\cdots\f$
 *
 * self[lsize] \f$\rightarrow\f$  x[start+lsize-1]
 *
 * @param self  A vector object.           
 * @param start The start index of x.      
 * @param x 	The attached vector object.
 * 
 * @return 0 on success.
 */
extern int parms_VecAttachVec(parms_Vec self, int start, parms_Vec x);

/** 
 * Attach vector self to the memory pointed to by array 
 * 
 * @param self  A vector                             
 * @param array The memory to which self is attached.
 * 
 * @return 0 on success.
 */
extern int parms_VecAttachArray(parms_Vec self, FLOAT *array);

/** 
 * Allocate the memory space for the parms_Vec object self.
 * 
 * @param self A vector object.
 * 
 * @return 0 on success.
 */
extern int parms_VecAlloc(parms_Vec self);

/** 
 * Perform permutation. 
 * 
 * aux[perm[i]] = self[i].
 *
 * @param self A vector object.                      
 * @param aux  An array which stores permuted values.
 * 
 * @return 0 on success.
 */
extern int parms_VecPermAux(parms_Vec self, FLOAT *aux);

/** 
 * Perform inverse permutation. 
 * 
 * self[i] = aux[perm[i]].
 *  
 * @param self A vector object.                               
 * @param aux  An working array which stores permuted values. 
 * 
 * @return 0 on success.
 */
extern int parms_VecInvPermAux(parms_Vec self, FLOAT *aux);

PARMS_CXX_END

#endif 
